package org.w3c.rdf.syntax.strawman;

import org.w3c.rdf.model.*;
import org.w3c.rdf.model.impl.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.RDFS;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.syntax.xml.*;
import org.w3c.rdf.util.*;
import org.w3c.tools.crypt.*;


import org.xml.sax.*;
import org.xml.sax.helpers.*;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * A parser for a simplified syntax for RDF. Supports arbitrary XML files.
 * Uses three control attributes:
 * <ol>
 * <li><code>rdf:instance</code>: specifes that a tag denotes an instance of a class</li>
 * <li><code>rdf:for</code>: specifes the subject URI of the property</li>
 * <li><code>rdf:resource</code>: specifes the object URI of the property</li>
 * </ol>
 *
 * @author    Sergey Melnik <melnik@db.stanford.edu>
 */

public class StrawmanParser extends GenericParser {

  public static final String REVISION = "Strawman RDF parser v0.1 1999-12-16";
  public static final String RDF_INSTANCE = RDF._Namespace + "instance";
  public static final String RDF_RESOURCE = RDF._Namespace + "resource";
  public static final String RDF_FOR = RDF._Namespace + "for";

  int shift = 0;
  byte[] pathDigest;

  public StrawmanParser () {

    super();
  }

  boolean isProperty(Element el) {

    //    StrawElement se = (StrawElement)el;
    return el.getAttribute(RDF_FOR) != null ||
      (el.getParent() != null /*((StrawElement)el).depth > 0*/ && el.getAttribute(RDF_INSTANCE) == null);
  }

  StrawElement getParent(Element e) {
    return (StrawElement)e.getParent();
  }

  Resource getArcFromParent(Element e) {
    return ((StrawElement)e).arcFromParent;
  }

  int depth(Element e) {
    return ((StrawElement)e).depth;
  }

  void collectCDATA(StrawElement p) {

    // check for deferred CDATA
    if(p.getValue().length() > 0 || (p.getType() == p.EMPTY && p.arcs.size() == 0 && p.getAttribute(RDF_RESOURCE) == null)) {
      p.addArc(p.next(), new LiteralImpl(p.getValue()));
      p.setValue(p.EMPTY_STR);
    }
  }

  Resource uri2resource(String uri) {

    if(uri.startsWith("#"))
      uri = getSourceURI() + uri;

    return new ResourceImpl(uri);
  }

  byte[] getDigest(Object obj) {

    if(obj instanceof Digestable)
      return ((Digestable)obj).getDigest().getDigestBytes();

    if(obj instanceof StrawElement) {

      StrawElement se = (StrawElement)obj;
      if(se.anonymous)
	return se.digestPath;
      else
	return getDigest(se.subject);
    }

    throw new RuntimeException("No digest for " + obj);
  }

  
  public void startElement (String name, AttributeList al) throws SAXException {
    
    _startElement(name, al);

    StrawElement current = (StrawElement)this.current;
    StrawElement parent = getParent(current);

    // set subject if possible. If anonymous do nothing
    String inst = current.getAttribute(RDF_INSTANCE);
    String res = current.getAttribute(RDF_RESOURCE);
    String fr = current.getAttribute(RDF_FOR);

    if(inst != null && res != null)
      throw new SAXException("Invalid RDF/XML element: rdf:instance and rdf:resource are mutually exclusive!");

    if(inst != null && inst.length() > 0)
      current.subject = uri2resource(inst);
    else if(res != null && res.length() > 0)
      current.subject = uri2resource(res);
    //    else if(fr != null && fr.length() > 0) // FIXME
    //      current.subject = uri2resource(fr);
    else
      current.anonymous = true;

    // collect non-RDF attributes
    Enumeration attrs = current.getAttributes();
    while(attrs.hasMoreElements()) {
      String aName = (String)attrs.nextElement();
      //      System.err.println("=== attr: " + aName);
      if(!(aName.equals(RDF_RESOURCE) || aName.equals(RDF_INSTANCE) || aName.equals(RDF_FOR)))
	current.addArc(uri2resource(aName), new LiteralImpl(current.getAttribute(aName)));
    }

    if(fr != null) {
      // node with rdf:for. Loose context. Set depth of this node to 0.
      
      // create an artificial node
      StrawElement el = new StrawElement();
      if(parent != null)
	parent.setChild(el);
      el.setChild(current);
      el.setAttribute(RDF_FOR, fr);
      current.setParent(el);
      el.setParent(parent);

      if(fr.length() > 0)
	el.subject = uri2resource(fr);
      else
	el.anonymous = true;


      //      current.setChild(el);
      //      el.setParent(current);
      //      current = el;
      
      current.arcFromParent = uri2resource(current.getName());
      el.addArc(current.arcFromParent, current);
      current.depth = 1;
      current.digestPath = getDigest(current.arcFromParent);

      //      el.arcFromParent = parent.next();
      //      parent.addArc(el.arcFromParent, el);

      //      if(res != null && res.length() > 0)
      //	current.subject = uri2resource(res);

    } else {
      // no rdf:for

      if(parent != null) {
	
	collectCDATA(parent);
	
	// set path from parent to current:
	// determine whether current is property or instance element
	if(isProperty(current)) {
	  current.arcFromParent = uri2resource(current.getName());
	} else {
	  // arc from parent is next ordinal
	  current.arcFromParent = parent.next();
	}
	parent.addArc(current.arcFromParent, current);

	current.depth = parent.depth + 1;
	// set path digest
	byte[] d = getDigest(getArcFromParent(current));
	if(depth(current) == 1) {
	  current.digestPath = d;
	} else {
	  //	  System.err.println("--Current: " + current.getName() + ", depth=" + current.depth);
	  current.digestPath = new byte[parent.digestPath.length];
	  System.arraycopy(parent.digestPath, 0, current.digestPath, 0, parent.digestPath.length);
	  DigestUtil.xor(current.digestPath, d, depth(current)-1);
	}
      }
    }

    //    System.out.println("Start element: " + current.getName());
  }


  public void endElement (String name) throws SAXException {

    finishElement();

    if(current.getAttribute(RDF_FOR) != null) {
      // remove the artificial node
      this.current = current.getParent();
      finishElement();
    }

    _endElement(name);
  }

  void finishElement() throws SAXException {

    //    System.out.println("End element: " + current.getName());

    StrawElement current = (StrawElement)this.current;
    boolean done = false;

    collectCDATA(current);

    // normalize arcs
    Vector arcs = current.arcs;
    // if only one arc _1 to a literal (no _2) replace it with rdf:value
    StrawArc firstArc = findArc(arcs, RDFUtil.ord(1));

    if(firstArc != null &&
       firstArc.object() instanceof Literal &&
       findArc(arcs, RDFUtil.ord(2)) == null) {

      firstArc.predicate = RDF.value;
    }

    if(arcs.size() == 1) {
      StrawArc singleArc = (StrawArc)arcs.elementAt(0);

      if(current.anonymous && singleArc.predicate().equals(RDF.value) && isProperty(current)) {

	// replace this node in parent's arc list by the literal
	//	System.err.println("===REPLACING in " + current.getName());
	Vector parcs = getParent(current).arcs;
	StrawArc parc = findArc(parcs, current);
	if(parc != null) {
	  //	      	    System.err.println("===FOUND at " + i);
	  // FIXME: singleArc.object must be a literal
	  parc.object = singleArc.object();
	  done = true;
	}
      }
    }

    if(!done) {

      if(!isProperty(current))
	current.addArc(RDF.type, uri2resource(current.getName()));
	//	createStatement(current.subject, RDF.type, uri2resource(current.getName()));

      if(current.anonymous) {
	// compute subject from digest
	for(int i=0; i < arcs.size(); i++) {
	  StrawArc arc = (StrawArc)arcs.elementAt(i);
	  updateDigest(current, arc);
	  //	System.out.println("    node " + current.getName() + " at depth " + current.depth + " adding " + arc.predicate() + "=" + arc.object());
	}
	current.subject = uri2resource(getSourceURI() + "#" + RDFDigestUtil.getDigestAlgorithm() +
				       "-" + DigestUtil.toHexString(current.digestPath));
      }
      
      // create triple for rdf:instance
      //    String inst = current.getAttribute(RDF_INSTANCE);
      
      // generate triples for the arcs
      for(int i = 0; i < arcs.size(); i++) {
	StrawArc arc = (StrawArc)arcs.elementAt(i);
	Object obj = arc.object();
	RDFNode object = obj instanceof RDFNode ? (RDFNode)obj : ((StrawElement)obj).subject;
	createStatement(current.subject, arc.predicate(), object);
      }
      // destroy arcs
      current.arcs = null;
    }

  }

  protected void updateDigest(StrawElement current, StrawArc arc) {

    if(current.digestPath == null) // i.e. depth == 0
      current.digestPath = getDigest(arc.predicate());
    else
      DigestUtil.xor(current.digestPath, getDigest(arc.predicate()), current.depth);
    DigestUtil.xor(current.digestPath, getDigest(arc.object()), current.depth + 1);
  }

  protected void createStatement(Resource subject, Resource predicate, RDFNode object) {

    if(!(predicate.equals(RDF.type) && object.equals(RDFS.Resource)))
       consumer.addStatement(subject, predicate, object);
  }

  StrawArc findArc(Vector arcs, Resource predicate) {

    for(int i = 0; i < arcs.size(); i++) {
      StrawArc arc = (StrawArc)arcs.elementAt(i);
      if(arc.predicate().equals(predicate))
	return arc;
    }
    return null;
  }

  StrawArc findArc(Vector arcs, StrawElement element) {

    for(int i = 0; i < arcs.size(); i++) {
      StrawArc arc = (StrawArc)arcs.elementAt(i);
      if(arc.object() == element)
	return arc;
    }
    return null;
  }

  public void characters (char ch[], int start, int length)
    throws SAXException {
  
      String s = new String (ch, start, length);
      String sTrimmed = s.trim();
      if (sTrimmed.length() > 0 || preserveWhiteSpace()) {
	//	System.err.println("---Setting " + current.getName() + " to " + sTrimmed);
	current.setValue(sTrimmed);
      }
  }

  static void bailOut() {
    System.err.println("Usage: java -Dorg.xml.sax.parser=<classname> org.w3c.rdf.syntax.strawman.StrawmanParser " +
		       //			 "[-fetch_schemas | -f] [-stream | -s] [-robust | -r] [ URI | filename ]");
		       "[ URI | filename ]");
    System.err.println ("This is revision " + REVISION);
    System.exit(1);
  }

  protected Element createElement() {

    return new StrawElement();
  }

  public static void main (String args[]) throws Exception {

    if(args.length != 1)
      bailOut();

    _main(args[0], new StrawmanParser());
  }

}
